from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

from phiterm.conf.constants import PHI_CONF_PATH
from phiterm.schemas.user import UserSchema
from phiterm.schemas.workspace import WorkspaceSchema
from phiterm.utils.cli_console import print_heading, print_info
from phiterm.utils.common import pickle_object_to_file, unpickle_object_from_file
from phiterm.utils.dttm import current_datetime_utc
from phiterm.utils.filesystem import delete_from_fs
from phiterm.utils.log import logger
from phiterm.workspace.phi_ws_data import PhiWsData
from phiterm.workspace.ws_enums import WorkspaceSetupActions


class PhiConf:
    """The PhiConf class manages user data for the phi cli"""

    def __init__(self, user_schema: Optional[UserSchema] = None) -> None:
        # Current user, populated after authenticating with the backend
        # To add a user, use the user setter
        self._user: Optional[UserSchema] = user_schema

        # Active ws - used as the default for all phi commands
        # To add an active workspace, use the active_ws_name setter
        self._active_ws_name: Optional[str] = None

        # ** The self._ws_data_map holds the mapping from ws_name to ws_data.
        self._ws_data_map: Dict[str, PhiWsData] = OrderedDict()
        # When phi is initialized using `phi init`, this is an empty dict
        # When the user is authenticated, we get a list of all the workspaces this user has access to.
        # These workspaces are then added to PhiConf using the `active_ws_name` and `available_ws` setters.
        # Note: These workspaces are not yet cloned on the users machines and
        # only a record of their availability is kept.
        # When the workspace is setup on the users machine, we update the remaining workspace data.

        # Quick access from ws_path -> ws data.
        self._path_to_ws_data_map: Dict[Path, PhiWsData] = OrderedDict()

        # Save the config to disk once initialized
        self.save_config()

    ######################################################
    ## User functions
    ######################################################

    @property
    def user(self) -> Optional[UserSchema]:
        return self._user

    @user.setter
    def user(self, user: Optional[UserSchema]) -> None:
        """Sets the user"""
        if user is not None:
            self._user = user
            self.save_config()

    ######################################################
    ## Workspace functions
    ######################################################

    @property
    def active_ws_name(self) -> Optional[str]:
        return self._active_ws_name

    @active_ws_name.setter
    def active_ws_name(self, ws_name: Optional[str]) -> None:
        if ws_name is not None:
            self._active_ws_name = ws_name
            self._add_or_update_ws_data(ws_name=ws_name)
            self.save_config()

    @property
    def available_ws(self) -> List[WorkspaceSchema]:
        return [
            w.ws_schema for w in self._ws_data_map.values() if w.ws_schema is not None
        ]

    @available_ws.setter
    def available_ws(self, avl_ws: Optional[List[WorkspaceSchema]]) -> None:
        if avl_ws:
            for ws_schema in avl_ws:
                self._add_or_update_ws_data(
                    ws_name=ws_schema.ws_name, ws_schema=ws_schema
                )
            self.save_config()

    def _add_or_update_ws_data(
        self,
        ws_name: Optional[str] = None,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_root_path: Optional[Path] = None,
        ws_config_file_path: Optional[Path] = None,
    ) -> None:
        """The main function to create, update or refresh PhiWsData in PhiConf.

        Notes:
        * ws_name is the only required argument.
        * This function does not call self.save_config(). call it from the parent.

        This function is called from 5 places:
        1. During `phi login` when active/available workspaces are requested from the api.
        2. During `phi ws init` when a new workspace is created on the users machine.
        3. When the user already has a local workspace on their machine
            and adds the ws manually using `phi ws setup`
            (possible after running `phi init -r` which deletes the config)
        4. Whenever ws config, creds or data needs to be refreshed. This function is called with just the `ws_name` arg
        5. When active or available workspaces are updated, this function is called from the
            @active_ws_name.setter and @available_ws.setter with the ws_name and ws_schema
        """

        ######################################################
        # Validate ws_name
        ######################################################

        # ws_name should not be None
        if ws_name is None or not isinstance(ws_name, str):
            return

        ######################################################
        # Create ws_data if it does not exist
        ######################################################

        if ws_name not in self._ws_data_map:
            logger.debug(f"Creating workspace: {ws_name}")
            self._ws_data_map[ws_name] = PhiWsData(
                ws_name=ws_name,
                ws_schema=ws_schema,
                ws_root_path=ws_root_path,
                ws_config_file_path=ws_config_file_path,
            )
            if ws_root_path is not None:
                self._path_to_ws_data_map[ws_root_path] = self._ws_data_map[ws_name]
                logger.debug(f"Workspace dir: {ws_root_path}")
            logger.debug(f"Workspace created: {ws_name}")
            return

        ######################################################
        # Update ws_data
        ######################################################

        logger.debug(f"Updating workspace: {ws_name}")
        # By this point there should be a PhiWsData object for this ws_name
        existing_ws_data: PhiWsData = self._ws_data_map.get(
            ws_name, PhiWsData(ws_name=ws_name)
        )
        # Make a deep copy of the existing workspace data
        # This allows us to update the fields we want, and keep any existing fields as is
        _ws_data: PhiWsData = existing_ws_data.copy(deep=True)

        # Update values where available
        if ws_schema is not None and ws_schema != existing_ws_data.ws_schema:
            logger.debug("Updating ws_schema")
            _ws_data.ws_schema = ws_schema
        if ws_root_path is not None and ws_root_path != existing_ws_data.ws_root_path:
            logger.debug("Updating ws_root_path")
            _ws_data.ws_root_path = ws_root_path
        if (
            ws_config_file_path is not None
            and ws_config_file_path != existing_ws_data.ws_config_file_path
        ):
            logger.debug("Updating ws_config_file_path")
            _ws_data.ws_config_file_path = ws_config_file_path
        _ws_data.last_update_ts = current_datetime_utc()

        # Point the ws_data in _ws_data_map and _path_to_ws_data_map to new _ws_data
        # 1. Pop the existing object from the self._ws_data_map
        if ws_name in self._ws_data_map:
            popped_ws_data_from_map: Optional[PhiWsData] = self._ws_data_map.pop(
                ws_name
            )
            # logger.debug(f"Removed {ws_name} from _ws_data_map")
        self._ws_data_map[ws_name] = _ws_data

        # 2. Pop the existing object from the self._path_to_ws_data_map
        if _ws_data.ws_root_path is not None:
            if (
                existing_ws_data.ws_root_path is not None
                and existing_ws_data.ws_root_path in self._path_to_ws_data_map
            ):
                popped_ws_data_from_path_map: Optional[
                    PhiWsData
                ] = self._path_to_ws_data_map.pop(existing_ws_data.ws_root_path)
                # logger.debug(
                #     f"Pop {existing_ws_data.ws_root_path} from _path_to_ws_data_map"
                # )
            self._path_to_ws_data_map[_ws_data.ws_root_path] = self._ws_data_map[
                ws_name
            ]
        logger.debug(f"Mapping updated for: {ws_name}")

    ######################################################
    # END
    ######################################################

    def add_new_ws_to_config(
        self,
        ws_name: str,
        ws_root_path: Path,
        ws_config_file_path: Path,
    ) -> None:
        """Adds a newly created workspace to the PhiConf"""
        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_root_path=ws_root_path,
            ws_config_file_path=ws_config_file_path,
        )
        self.save_config()

    def update_ws_data(
        self,
        ws_name: str,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_root_path: Optional[Path] = None,
        ws_config_file_path: Optional[Path] = None,
        set_as_active: bool = False,
    ) -> None:
        """Update workspace data and return True if successful"""

        self._add_or_update_ws_data(
            ws_name=ws_name,
            ws_schema=ws_schema,
            ws_root_path=ws_root_path,
            ws_config_file_path=ws_config_file_path,
        )
        if set_as_active:
            self.active_ws_name = ws_name
        self.save_config()

    def validate_workspace(
        self, ws_name: str
    ) -> Tuple[bool, Optional[Set[WorkspaceSetupActions]]]:
        """Validates a workspace and returns a tuple of [validation_status, pending_setup_actions]
        Returns:
            (False, None): If PhiWsData is not valid
            (True, None): If Validation is successful and everything is properly setup
            (True, set(WorkspaceSetupActions)): If Validation is successful but user has pending setup actions
            (False, set(WorkspaceSetupActions)): If Validation is unsuccessful and user has pending setup actions
        """

        # Validate the ws has a valid entry
        if ws_name is None or ws_name not in self._ws_data_map:
            return False, None
        logger.debug(f"Validating workspace: {ws_name}")

        ws_data: PhiWsData = self._ws_data_map[ws_name]

        # If ws_config is missing add a pending action
        if ws_data.ws_config is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.WS_CONFIG_IS_AVL)
        else:
            ws_data.pending_actions.discard(WorkspaceSetupActions.WS_CONFIG_IS_AVL)

        # If ws_schema is missing add a pending action
        if ws_data.ws_schema is None:
            ws_data.pending_actions.add(WorkspaceSetupActions.WS_IS_AUTHENTICATED)
        else:
            ws_data.pending_actions.discard(WorkspaceSetupActions.WS_IS_AUTHENTICATED)

        # If git_url is missing add a pending action
        if ws_data.ws_schema is None:
            from phiterm.utils.git import get_remote_origin_for_dir

            _remote_origin = get_remote_origin_for_dir(ws_data.ws_root_path)
            if _remote_origin is None:
                ws_data.pending_actions.add(
                    WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL
                )
        # elif ws_data.ws_schema.git_url is None:
        #     ws_data.pending_actions.add(WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL)
        # else:
        #     ws_data.pending_actions.discard(
        #         WorkspaceSetupActions.GIT_REMOTE_ORIGIN_IS_AVL
        #     )

        # logger.debug(f"Required Actions for this workspace: {ws_data.required_actions}")
        # logger.debug(f"Pending Actions for this workspace: {ws_data.pending_actions}")

        # Check how many required_actions are pending
        if len(ws_data.required_actions.intersection(ws_data.pending_actions)) > 0:
            return False, ws_data.pending_actions

        return True, ws_data.pending_actions

    def delete_ws(self, ws_name: str, delete_ws_dir: bool = False) -> bool:
        """Handles Deleting a workspace from the PhiConf"""

        print_info(f"Deleting record for: {ws_name}")
        print_info(
            f"-*- Note: this does not delete any files on disk, please delete them manually"
        )

        ws_data: Optional[PhiWsData] = self._ws_data_map.pop(ws_name, None)
        if ws_data is None:
            logger.error(f"No record of workspace {ws_name}")
            return False

        ws_data_from_path: Optional[PhiWsData]
        if ws_data.ws_root_path is not None:
            ws_data_from_path = self._path_to_ws_data_map.pop(
                ws_data.ws_root_path, None
            )

        # Check if we're deleting the active workspace, if yes, unset the active ws
        if (
            self._active_ws_name is not None
            and ws_data.ws_name is not None
            and self._active_ws_name == ws_data.ws_name
        ):
            print_info(f"Removing {ws_data.ws_name} as the active workspace")
            self._active_ws_name = None

        ws_root_path_deleted: bool = False
        if delete_ws_dir and ws_data.ws_root_path:
            print_info(f"Deleting {str(ws_data.ws_root_path)}")
            ws_root_path_deleted = delete_from_fs(ws_data.ws_root_path)

        # Save the config after deleting workspace
        self.save_config()
        return True

    ######################################################
    ## Get Workspace Data
    ######################################################

    def get_ws_data_by_name(self, ws_name: str) -> Optional[PhiWsData]:
        return self._ws_data_map[ws_name] if ws_name in self._ws_data_map else None

    def get_ws_data_by_path(self, ws_root_path: Path) -> Optional[PhiWsData]:
        return (
            self._path_to_ws_data_map[ws_root_path]
            if ws_root_path in self._path_to_ws_data_map
            else None
        )

    ## Workspace Name
    def get_ws_name_by_path(self, ws_root_path: Path) -> Optional[str]:
        if ws_root_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_root_path].ws_name
        return None

    ## Workspace Schema
    def get_ws_schema_by_name(self, ws_name: str) -> Optional[WorkspaceSchema]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_schema
        return None

    def get_ws_schema_by_path(self, ws_root_path: Path) -> Optional[WorkspaceSchema]:
        if ws_root_path in self._path_to_ws_data_map:
            return self._path_to_ws_data_map[ws_root_path].ws_schema
        return None

    ## Workspace Directory
    def get_ws_root_path_by_name(self, ws_name: str) -> Optional[Path]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_root_path
        return None

    ## Workspace Config File Path
    def get_ws_config_file_path_by_name(self, ws_name: str) -> Optional[Path]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].ws_config_file_path
        return None

    ## Workspace Pending Actions
    def get_ws_pending_actions_by_name(
        self, ws_name: str
    ) -> Optional[Set[WorkspaceSetupActions]]:
        if ws_name in self._ws_data_map:
            return self._ws_data_map[ws_name].pending_actions
        return None

    ## Workspace Data
    def get_active_ws_data(self, refresh: bool = False) -> Optional[PhiWsData]:
        if self.active_ws_name is not None and self.active_ws_name in self._ws_data_map:
            if refresh:
                self.refresh_ws_config(self.active_ws_name)
            return self._ws_data_map[self.active_ws_name]
        return None

    def refresh_ws_config(self, ws_name: str) -> None:
        """Refresh the workspace config for a given workspace name"""

        self._add_or_update_ws_data(ws_name=ws_name)
        logger.debug(f"++**++ Config refreshed: {ws_name}")
        self.save_config()

    ######################################################
    ## Sync Workspace Data from api
    ######################################################

    def sync_workspaces_from_api(self) -> bool:
        from phiterm.api.workspace import (
            get_primary_workspace,
            get_available_workspaces,
        )

        if self.user is None:
            logger.error("User not available for workspace sync")
            return False

        # If the active_ws_name is None, call backend to get the primary_ws
        primary_ws: Optional[WorkspaceSchema] = None
        if self.active_ws_name is None:
            primary_ws = get_primary_workspace(self.user)
            if primary_ws is None:
                logger.debug("No primary workspace available")
            else:
                logger.debug(f"Received primary ws: {primary_ws.json(indent=2)}")
                self.active_ws_name = primary_ws.ws_name

        # If available_workspaces is empty and a primary_ws is available
        # call backend to get the available workspaces
        available_workspaces: Optional[List[WorkspaceSchema]] = None
        if len(self.available_ws) == 0:
            available_workspaces = get_available_workspaces(self.user)
            if available_workspaces is not None:
                logger.debug(
                    f"Received {len(available_workspaces)} available workspaces"
                )
                self.available_ws = available_workspaces
        return True

    ######################################################
    ## PhiConf functions
    ######################################################

    def save_config(self):
        # logger.debug(f"Saving config to {str(PHI_CONF_PATH)}")
        pickle_object_to_file(self, PHI_CONF_PATH)

    @classmethod
    def get_saved_conf(cls):
        # logger.debug(f"Reading phidata conf at {PHI_CONF_PATH}")
        if PHI_CONF_PATH.exists():
            # logger.debug(f"{PHI_CONF_PATH} exists")
            if PHI_CONF_PATH.is_file():
                return unpickle_object_from_file(
                    file_path=PHI_CONF_PATH, verify_class=cls
                )
            elif PHI_CONF_PATH.is_dir():
                logger.debug(f"{PHI_CONF_PATH} is a directory, deleting and exiting")
                delete_from_fs(PHI_CONF_PATH)
        return None

    def print_to_cli(self):
        if self.user:
            print_heading(f"User: {self.user.email}")
        if self.active_ws_name:
            print_heading(f"\nActive workspace: {self.active_ws_name}\n")
        else:
            print_info("* No active workspace, run `phi ws init` to create one")
        if len(self._ws_data_map) > 0:
            print_heading("Available workspaces:\n")
            c = 1
            for k, v in self._ws_data_map.items():
                print_info(f"  {c}. Name: {v.ws_name}")
                print_info(f"     Dir: {v.ws_root_path}")
                if v.ws_schema:
                    logger.debug(
                        f"     Schema: {v.ws_schema.json(exclude_none=True, indent=2)}"
                    )
                print_info(f"  -*-\n")
                c += 1

    ######################################################
    ## Deprecated functions
    ######################################################
    #
    # def add_ws_data_using_config_file(
    #     self, ws_root_path: Path, ws_config_file_path: Optional[Path]
    # ) -> Optional[PhiWsData]:
    #     """Maps the workspace at `ws_root_path` to an
    #     available workspace using the `ws_config_file_path`.
    #     Also updates the ws_config_file_path in the ws_data if provided.
    #
    #     This is used in 2 cases:
    #     1. If the user ran `phi init -r`, the PhiConf gets erased and so does self._path_to_ws_data_map
    #         So there is no link from the ws_root_path to the workspace.
    #     2. The user manually cloned the workspace directory. In that case we don't have a record of this
    #         ws_root_path matching any available workspace.
    #
    #     This function basically reads the workspace config and tries to map it to an available workspace.
    #     Returns the PhiWsData if successful
    #     """
    #
    #     logger.debug(f"Looking for the workspace at {ws_root_path}")
    #     _best_guess_ws_name = ws_root_path.stem
    #     logger.debug(f"_best_guess_ws_name: {_best_guess_ws_name}")
    #     _ws_schema: Optional[WorkspaceSchema] = None
    #     for avl_ws in self.available_ws:
    #         if avl_ws.ws_name == _best_guess_ws_name:
    #             _ws_schema = avl_ws
    #         break
    #     if _ws_schema is None:
    #         logger.error(
    #             f"Workspace name: {_best_guess_ws_name} does not match any available workspaces"
    #         )
    #         print_available_workspaces(self.available_ws)
    #         return None
    #
    #     logger.debug(f"Found matching workspace {_ws_schema.ws_name}")
    #     logger.debug(
    #         f"Mapping {_ws_schema.ws_name} to dir: {ws_root_path} and config: {ws_config_file_path}"
    #     )
    #     self._add_or_update_ws_data(
    #         ws_name=_ws_schema.ws_name,
    #         ws_schema=_ws_schema,
    #         ws_root_path=ws_root_path,
    #         ws_config_file_path=ws_config_file_path,
    #     )
    #     self.save_config()
    #     return self._ws_data_map.get(_ws_schema.ws_name, None)



import logging
import sys

from phiterm.conf.constants import PHI_LOGGER_NAME, PHIDATA_LOGGER_NAME


def get_simple_logger(logger_name: str) -> logging.Logger:

    handler = logging.StreamHandler(sys.stdout)
    # handler.setLevel(logging.INFO)
    simple_formatter = logging.Formatter(
        "%(lineno)d - %(filename)s - %(levelname)s - %(message)s"
    )
    # dttm_formatter = logging.Formatter(
    #     "%(asctime)s - %(lineno)d - %(filename)s - %(levelname)s - %(message)s",
    #     "%Y-%m-%d %H:%M",
    # )
    handler.setFormatter(simple_formatter)

    _logger = logging.getLogger(logger_name)
    _logger.addHandler(handler)
    _logger.setLevel(logging.INFO)
    # _logger.propagate = False
    return _logger


def get_rich_logger(logger_name: str) -> logging.Logger:

    from rich.logging import RichHandler

    # https://rich.readthedocs.io/en/latest/reference/logging.html#rich.logging.RichHandler
    # https://rich.readthedocs.io/en/latest/logging.html#handle-exceptions
    rich_handler = RichHandler(
        show_time=False, rich_tracebacks=False, tracebacks_show_locals=False
    )
    rich_handler.setFormatter(
        logging.Formatter(
            fmt="%(message)s",
            datefmt="[%X]",
        )
    )

    _logger = logging.getLogger(logger_name)
    _logger.addHandler(rich_handler)
    _logger.setLevel(logging.INFO)
    # _logger.propagate = False
    return _logger


logger: logging.Logger = get_rich_logger(PHI_LOGGER_NAME)


def set_log_level_to_debug():
    phi_logger = logging.getLogger(PHI_LOGGER_NAME)
    phi_logger.setLevel(logging.DEBUG)
    phidata_logger = logging.getLogger(PHIDATA_LOGGER_NAME)
    phidata_logger.setLevel(logging.DEBUG)



# import datetime
# from pathlib import Path
# from typing import Optional
#
# from pydantic import BaseModel
#
# from pak8.k8s.resources.kubeconfig import Kubeconfig
# from phiterm.conf.constants import PHI_KUBE_DIR
# from phiterm.utils.dttm import current_datetime_utc
# from phiterm.utils.log import logger
#
#
# class PhiK8sData(BaseModel):
#
#     # Set to True if the kubeconfig is available as a resource or in a file
#     # This flag saves us reading a file if the kubeconfig isnt available
#     kubeconfig_avl: bool = False
#     kubeconfig_resource: Optional[Kubeconfig] = None
#     kubeconfig_path: Optional[Path] = None
#     last_update_ts: datetime.datetime = current_datetime_utc()
#
#     @classmethod
#     def create_using_kubeconfig_resource(
#         cls, ws_name: str, kconf_resource: Kubeconfig
#     ) -> Optional["PhiK8sData"]:
#
#         if ws_name is None or kconf_resource is None:
#             return None
#
#         if not PHI_KUBE_DIR.exists():
#             PHI_KUBE_DIR.mkdir(exist_ok=True)
#
#         logger.debug("Creating PhiK8sData")
#         kconf_path: Path = PHI_KUBE_DIR.joinpath(f"{ws_name}_kubeconfig")
#         logger.debug(f"Saving kubeconfig to {str(kconf_path)}")
#         kconf_path.write_text(kconf_resource.json(by_alias=True))
#
#         return cls(
#             kubeconfig_avl=True,
#             kubeconfig_resource=kconf_resource,
#             kubeconfig_path=kconf_path,
#         )
#
#     def update_kubeconfig_resource(self, kconf_resource: Kubeconfig) -> bool:
#
#         if kconf_resource is None:
#             return False
#
#         logger.debug(f"Saving kubeconfig to {str(self.kubeconfig_path)}")
#         self.kubeconfig_resource = kconf_resource
#         if self.kubeconfig_path:
#             self.kubeconfig_path.write_text(kconf_resource.json(by_alias=True))
#             return True
#
#         return False
#
#     def get_kubeconfig(self) -> Optional[Kubeconfig]:
#         if not self.kubeconfig_avl:
#             logger.debug("No Kubeconfig available")
#             return None
#         # TODO: Update this to read from kubeconfig_path if kubeconfig_resource is None
#         return self.kubeconfig_resource



# from typing import Optional
#
# from phi import schemas
# from phiterm.conf.phi_conf import PhiGcpData, PhiWsData
# from phiterm.utils.common import pprint_info, pprint_info
# from phiterm.utils.log import logger
#
#
# def secho_gcp_status(ws_data: PhiWsData) -> None:
#     ws_gcp_data: Optional[PhiGcpData] = ws_data.ws_gcp_data
#     if ws_gcp_data is None:
#         pprint_info("")
#         pprint_info(
#             "No GCP Project set up, run `phi gcp auth` to setup your GCP Project"
#         )
#         return None
#
#     # pprint_heading("GCP")
#
#     gcp_project: Optional[schemas.GCPProjectSchema] = ws_gcp_data.gcp_project
#     # if gcp_project:
#     #     gcp_project.secho_status()
#
#     gke_cluster: Optional[schemas.GKEClusterSchema] = ws_gcp_data.gke_cluster
#     # if gke_cluster:
#     #     gke_cluster.secho_status()
#
#     gcloud_default_creds_avl = ws_gcp_data.gcloud_default_creds_avl
#     # gcp_svc_account: Optional[Dict[str, Any]] = ws_gcp_data.gcp_svc_account
#     # gcp_svc_account_key: Optional[Dict[str, Any]] = ws_gcp_data.gcp_svc_account_key
#     logger.debug(
#         "Gcloud Default Credentials Available: {}".format(gcloud_default_creds_avl)
#     )
#     # logger.debug("gcp_svc_account: {}".format(gcp_svc_account))
#     # logger.debug("gcp_svc_account_key: {}".format(gcp_svc_account_key))



from typing import Any, Dict, Optional

from pydantic import BaseModel


class WorkspaceSchema(BaseModel):
    """
    Schema for user workspace used by API.
    """

    ws_name: Optional[str] = None
    id_user: Optional[int] = None
    id_workspace: Optional[int] = None
    is_primary_ws_for_user: Optional[bool] = None
    visibility: Optional[str] = None
    ws_data: Optional[Dict[str, Any]] = None
    is_active: Optional[bool] = None
    is_test: Optional[bool] = None


class WorkspaceActionData(BaseModel):
    """
    Schema for workspace action data.
    """

    id_user: int
    id_workspace: int
    action: str
    action_data: Optional[Dict[str, Any]] = None



from os import environ
from pathlib import Path
from typing import Optional, Dict, List

from phidata import constants
from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.task.task import Task
from phidata.workspace import WorkspaceConfig
from phidata.types.context import PathContext, RunContext

from phiterm.types.run_status import RunStatus
from phiterm.utils.cli_console import (
    print_info,
    print_heading,
    print_dict,
    print_info,
    print_subheading,
)
from phiterm.utils.log import logger
from phiterm.workspace.phi_ws_data import PhiWsData


def set_local_env(
    ws_config: WorkspaceConfig,
    path_context: PathContext,
) -> None:
    """VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    This function updates the local environment with Paths and Configuration
    Main uses:
        1. Path env variables used by DataAssets for building file_paths
        2. Cloud config variables used by AwsAssets like aws_region and profile
    """
    # 1. Path env variables used by DataAssets for building file_paths
    if path_context.scripts_dir is not None:
        environ[constants.SCRIPTS_DIR_ENV_VAR] = str(path_context.scripts_dir)
    if path_context.storage_dir is not None:
        environ[constants.STORAGE_DIR_ENV_VAR] = str(path_context.storage_dir)
    if path_context.meta_dir is not None:
        environ[constants.META_DIR_ENV_VAR] = str(path_context.meta_dir)
    if path_context.products_dir is not None:
        environ[constants.PRODUCTS_DIR_ENV_VAR] = str(path_context.products_dir)
    if path_context.notebooks_dir is not None:
        environ[constants.NOTEBOOKS_DIR_ENV_VAR] = str(path_context.notebooks_dir)
    if path_context.workspace_config_dir is not None:
        environ[constants.WORKSPACE_CONFIG_DIR_ENV_VAR] = str(
            path_context.workspace_config_dir
        )
    # 2. Cloud config variables used by AwsAssets like aws_region and profile
    if ws_config.aws_region is not None:
        environ[constants.AWS_REGION_ENV_VAR] = ws_config.aws_region
        environ[constants.AWS_DEFAULT_REGION_ENV_VAR] = ws_config.aws_region
    if ws_config.aws_profile is not None:
        environ[constants.AWS_PROFILE_ENV_VAR] = ws_config.aws_profile
    if ws_config.aws_config_file is not None:
        environ[constants.AWS_CONFIG_FILE_ENV_VAR] = ws_config.aws_config_file
    if ws_config.aws_shared_credentials_file is not None:
        environ[
            constants.AWS_SHARED_CREDENTIALS_FILE_ENV_VAR
        ] = ws_config.aws_shared_credentials_file
    if ws_config.local_env:
        environ.update(ws_config.local_env)


def run_workflows_local(
    workflow_file: str,
    run_context: RunContext,
    ws_data: PhiWsData,
    workflows: Dict[str, Workflow],
    workflow_name: Optional[str] = None,
    tasks: Optional[Dict[str, Task]] = None,
    task_name: Optional[str] = None,
) -> bool:
    """
    Runs a Workflow/Task in the local environment

    Args:
        workflow_file: Path of the workflow file relative to the products_dir.
            This is used to build the path_context
        run_context: RunContext
        ws_data: PhiWsData
        workflows: Dict[str, Workflow] to run
        workflow_name: Name of workflow to run
        tasks:
        task_name: Name of task to run

    Returns:
        run_status (bool): True is the Workflow ran successfully
    """
    from phidata.utils.cli_console import print_run_status_table

    # Validate
    if ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return False
    if ws_data.ws_root_path is None:
        logger.error("Workspace directory invalid")
        return False

    # Step 1: Build the PathContext
    ws_config = ws_data.ws_config
    ws_root_path = ws_data.ws_root_path
    scripts_dir_path = ws_root_path.joinpath(ws_config.scripts_dir).resolve()
    storage_dir_path = ws_root_path.joinpath(ws_config.storage_dir).resolve()
    meta_dir_path = ws_root_path.joinpath(ws_config.meta_dir).resolve()
    products_dir_path = ws_root_path.joinpath(ws_config.products_dir).resolve()
    notebooks_dir_path = ws_root_path.joinpath(ws_config.notebooks_dir).resolve()
    workspace_config_dir_path = ws_root_path.joinpath(
        ws_config.workspace_config_dir
    ).resolve()
    workflow_file_path = Path(products_dir_path).joinpath(workflow_file)
    path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_path,
        storage_dir=storage_dir_path,
        meta_dir=meta_dir_path,
        products_dir=products_dir_path,
        notebooks_dir=notebooks_dir_path,
        workspace_config_dir=workspace_config_dir_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {path_context.json(indent=4)}")

    ######################################################################
    # Step 2: Update the local environment with Paths and Configuration
    # NOTE: VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    ######################################################################
    set_local_env(ws_config=ws_config, path_context=path_context)

    logger.debug("Running:")
    logger.debug(f"\tWorkflow: {workflow_name}")
    logger.debug(f"\tTask: {task_name}")

    # Step 3: If workflow_name or task_name is provided,
    # run single workflow or task
    if workflow_name is not None or task_name is not None:
        # workflow_name != "" means run a workflow
        if workflow_name is not None and workflow_name != "":
            workflow_to_run: Optional[Workflow] = workflows.get(workflow_name, None)
            if workflow_to_run is None:
                logger.error(
                    "Workflow `{}` not found in {}".format(
                        workflow_name, "[{}]".format(", ".join(workflows.keys()))
                    )
                )
                return False

            _name = workflow_to_run.name
            print_subheading(f"\nRunning Workflow: {_name}")
            # Pass down context
            workflow_to_run.run_context = run_context
            workflow_to_run.path_context = path_context
            wf_run_success = workflow_to_run.run_in_local_env(task_id=task_name)
            return wf_run_success

        # workflow_name == "" and task_name != "" means run a task
        elif (
            workflow_name == ""
            and task_name is not None
            and task_name != ""
            and tasks is not None
        ):
            task_to_run: Optional[Task] = tasks.get(task_name, None)
            if task_to_run is None:
                logger.error(
                    "Task `{}` not found in {}".format(
                        task_name, "[{}]".format(", ".join(tasks.keys()))
                    )
                )
                return False

            _name = task_to_run.name
            print_subheading(f"\nRunning Task: {_name}")
            # Pass down context
            task_to_run.run_context = run_context
            task_to_run.path_context = path_context
            task_run_success = task_to_run.run_in_local_env()
            return task_run_success

    # Step 4: If workflow_name is None and task_name is None
    # means this is a file with just workflows or tasks
    # and no data products.
    # Run all workflows or tasks
    if workflows is not None and len(workflows) > 0:
        wf_run_status: List[RunStatus] = []
        for wf_name, wf_obj in workflows.items():
            _name = wf_name or wf_obj.name
            print_subheading(f"\nRunning Workflow: {_name}")
            # Pass down context
            wf_obj.run_context = run_context
            wf_obj.path_context = path_context
            wf_run_success = wf_obj.run_in_local_env()
            wf_run_status.append(RunStatus(name=_name, success=wf_run_success))

        print_run_status_table("Workflow Run Status", wf_run_status)
        for run_status in wf_run_status:
            if not run_status.success:
                return False
    elif tasks is not None and len(tasks) > 0:
        task_run_status: List[RunStatus] = []
        for task_name, task_obj in tasks.items():
            _name = task_name or task_obj.name
            print_subheading(f"\nRunning Task: {_name}")
            # Pass down context
            task_obj.run_context = run_context
            task_obj.path_context = path_context
            task_run_success = task_obj.run_in_local_env()
            task_run_status.append(RunStatus(name=_name, success=task_run_success))

        print_run_status_table("Task Run Status", task_run_status)
        for run_status in task_run_status:
            if not run_status.success:
                return False
    return True


def run_data_products_local(
    workflow_file: str,
    data_products: Dict[str, DataProduct],
    run_context: RunContext,
    ws_data: PhiWsData,
) -> bool:
    """
    Runs a DataProduct in the local environment

    Args:
        workflow_file: Path of the workflow file relative to the products_dir.
            This is used to build the path_context
        data_products: Dict[str, DataProduct] to run
        run_context: RunContext
        ws_data: PhiWsData

    Returns:
        run_status (bool): True is the DataProduct ran successfully
    """
    from phidata.utils.cli_console import print_run_status_table

    # Validate
    if ws_data.ws_root_path is None:
        logger.error("Workspace directory invalid")
        return False
    if ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return False

    # Step 1: Build the PathContext for the DataProducts.
    ws_config = ws_data.ws_config
    ws_root_path = ws_data.ws_root_path
    scripts_dir_path = ws_root_path.joinpath(ws_config.scripts_dir).resolve()
    storage_dir_path = ws_root_path.joinpath(ws_config.storage_dir).resolve()
    meta_dir_path = ws_root_path.joinpath(ws_config.meta_dir).resolve()
    products_dir_path = ws_root_path.joinpath(ws_config.products_dir).resolve()
    notebooks_dir_path = ws_root_path.joinpath(ws_config.notebooks_dir).resolve()
    workspace_config_dir_path = ws_root_path.joinpath(
        ws_config.workspace_config_dir
    ).resolve()
    workflow_file_path = Path(products_dir_path).joinpath(workflow_file)
    dp_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_path,
        storage_dir=storage_dir_path,
        meta_dir=meta_dir_path,
        products_dir=products_dir_path,
        notebooks_dir=notebooks_dir_path,
        workspace_config_dir=workspace_config_dir_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {dp_path_context.json(indent=4)}")

    ######################################################################
    # Step 2: Update the local environment with Paths and Configuration
    # NOTE: VERY IMPORTANT TO RUN BEFORE RUNNING ANY LOCAL WORKFLOWS
    ######################################################################
    set_local_env(ws_config=ws_config, path_context=dp_path_context)

    # Step 3: Run the DataProducts
    dp_run_status: List[RunStatus] = []
    for dp_name, dp_obj in data_products.items():
        _name = dp_name or dp_obj.name
        print_subheading(f"\nRunning DataProduct: {_name}")
        # Pass down context
        dp_obj.run_context = run_context
        dp_obj.path_context = dp_path_context
        run_success = dp_obj.run_in_local_env()
        dp_run_status.append(RunStatus(name=_name, success=run_success))

    print_run_status_table("DataProduct Run Status", dp_run_status)
    for run_status in dp_run_status:
        if not run_status.success:
            return False
    return True



"""Phi Workflow Cli

This is the entrypoint for the `phi wf` commands
"""

from datetime import datetime
from typing import Optional, cast

import typer

from phiterm.utils.cli_console import (
    print_conf_not_available_msg,
    print_active_workspace_not_available,
    print_available_workspaces,
)
from phiterm.workflow.wf_enums import WorkflowEnv
from phiterm.utils.dttm import dttm_str_to_dttm
from phiterm.utils.log import logger, set_log_level_to_debug

wf_app = typer.Typer(
    name="wf",
    short_help="Run workflows",
    help="""\b
Use `phi wf <command>` to run workflows.
Run `phi wf <command> --help` for more info.
    """,
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="<command>",
)


@wf_app.command(short_help="Run workflow", options_metavar="\b", no_args_is_help=True)
def run(
    workflow: str = typer.Argument(
        ...,
        help="Workflow path: `dir/file` inside the workflow directory",
        metavar="[workflow]",
        show_default=False,
    ),
    env_filter: Optional[str] = typer.Option(
        "local",
        "-e",
        "--env",
        metavar="",
        help="The environment to run the workflow in. Default: local",
        show_default=False,
    ),
    app_name: Optional[str] = typer.Option(
        None,
        "-a",
        metavar="",
        help="The App to run the workflow in.",
        show_default=False,
        hidden=True,
    ),
    run_date: Optional[str] = typer.Option(
        None,
        "-dt",
        "--date",
        help="Run datetime for the workflow. Example: '2023-01-01' or '2023-01-01T01:00'",
        show_default=False,
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Perform a dry run for each task. Does not run the task.",
    ),
    detach: bool = typer.Option(
        False,
        "--detach",
        help="[experimental] Run the command in the background.",
        hidden=True,
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Run a workflow in the target environment.

    \b
    Examples:
    $ phi wf run crypto/prices -> Runs the crypto/prices workflow in the local environment
    """
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.workflow.wf_operator import run_workflow
    from phiterm.utils.load_env import load_env

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data()
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(dotenv_dir=active_ws_data.ws_root_path)

    target_env: Optional[WorkflowEnv] = None
    target_app: Optional[str] = app_name
    target_dttm: Optional[datetime] = (
        dttm_str_to_dttm(run_date) if run_date else datetime.now()
    )
    if target_dttm is None:
        logger.error("Invalid run_date")
        return

    default_env = (
        active_ws_data.ws_config.default_env if active_ws_data.ws_config else None
    )
    target_env_str = env_filter or default_env or "dev"
    try:
        target_env = cast(WorkflowEnv, WorkflowEnv.from_str(target_env_str))
    except Exception as e:
        logger.error(e)
        logger.error(
            f"{target_env_str} is not supported, please choose from: {WorkflowEnv.values_list()}"
        )
        return

    logger.debug("Running workflow")
    logger.debug(f"\tworkflow     : {workflow}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_dttm  : {target_dttm}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tdetach       : {detach}")
    logger.debug(f"\ttarget_app   : {target_app}")
    run_workflow(
        wf_description=workflow,
        ws_data=active_ws_data,
        target_env=target_env,
        target_dttm=target_dttm,
        dry_run=dry_run,
        detach=detach,
        target_app=target_app,
    )


# @wf_app.command(short_help="Build workflow")
# def build(workflow: str = typer.Argument(..., help="The workflow to build")):
#     print_info(f"Building workflow {workflow}")
#     print_info(f"To be implemented")
#


@wf_app.command(short_help="List workflows", options_metavar="\b", no_args_is_help=True)
def ls(
    workflow: str = typer.Argument(
        ...,
        help="Workflow path: `dir/file` inside the workflow directory",
        metavar="[workflow]",
        show_default=False,
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.workflow.wf_operator import list_workflows
    from phiterm.utils.load_env import load_env

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(dotenv_dir=active_ws_data.ws_root_path)

    list_workflows(
        wf_description=workflow,
        ws_data=active_ws_data,
    )



# from typing import Any, Dict, List, Optional, Union
#
# import httpx
#
# from pak8.conf import Pak8Conf
# from phi import exceptions, schemas
# from phiterm.utils.common import pprint_error, pprint_network_error
# from phiterm.utils.log import logger
# from phiterm.zeus_api.client import ZeusApi, get_authenticated_client, validate_response
#
#
# def convert_gcp_project_response_to_schema(
#     resp: httpx.Response,
# ) -> Optional[schemas.GCPProjectSchema]:
#
#     gcp_project_dict = resp.json()
#     if gcp_project_dict is None:
#         return None
#     if not isinstance(gcp_project_dict, dict):
#         raise exceptions.ZeusGcpApiException(
#             "Could not parse GCPProjectSchema response"
#         )
#
#     # If the request failed, display the message from ApiResponseData
#     if (
#         resp.status_code == httpx.codes.BAD_REQUEST
#         and gcp_project_dict.get("status", "FAIL") == "FAIL"
#     ):
#         pprint_error(
#             gcp_project_dict.get("message", gcp_project_dict.get("detail", "FAIL"))
#         )
#
#     # If everything is good, convert gcp_project_dict to GCPProjectSchema and return
#     if resp.status_code == httpx.codes.OK:
#         return schemas.GCPProjectSchema.from_dict(gcp_project_dict)
#
#     return None
#
#
# def get_primary_gcp_project(
#     id_workspace: int,
# ) -> Optional[schemas.GCPProjectSchema]:
#
#     logger.debug("Getting active GCPProjectSchema")
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.get(
#                 "{}/{}".format(ZeusApi.read_primary_gcp_project, id_workspace)
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         # logger.debug("url: {}".format(r.url))
#         # logger.debug("status: {}".format(r.status_code))
#         # logger.debug("json: {}".format(r.json()))
#         return convert_gcp_project_response_to_schema(r)
#
#
# def upsert_gcp_project(
#     gcp_project_upsert: schemas.UpsertGCPProjectFromCli,
# ) -> Optional[schemas.GCPProjectSchema]:
#
#     logger.debug(
#         "Upserting GCPProjectSchema. Post: {}".format(ZeusApi.upsert_gcp_project)
#     )
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.post(
#                 ZeusApi.upsert_gcp_project,
#                 data=gcp_project_upsert.dict(),
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         # logger.debug("url: {}".format(r.url))
#         # logger.debug("status: {}".format(r.status_code))
#         # logger.debug("json: {}".format(r.json()))
#
#         if r.status_code == httpx.codes.OK:
#             logger.debug("GCPProjectSchema Upserted")
#             return convert_gcp_project_response_to_schema(r)
#
#     return None
#
#
# def update_gcp_project(
#     gcp_project_update: schemas.UpdateGCPProject,
# ) -> Optional[schemas.GCPProjectSchema]:
#
#     logger.debug(
#         "Updating GCPProjectSchema. Post: {}".format(ZeusApi.update_gcp_project)
#     )
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.post(
#                 ZeusApi.update_gcp_project, data=gcp_project_update.dict()
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         # logger.debug("url: {}".format(r.url))
#         # logger.debug("status: {}".format(r.status_code))
#         # logger.debug("json: {}".format(r.json()))
#         return convert_gcp_project_response_to_schema(r)
#
#
# def get_gke_cluster_for_gcp_project(
#     id_workspace: int,
#     id_project: str,
# ) -> Optional[schemas.GKEClusterSchema]:
#
#     logger.debug(
#         "Getting GKEClusterSchema for workspace {} and project {}".format(
#             id_workspace, id_project
#         )
#     )
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             r: httpx.Response = api.get(
#                 ZeusApi.get_gke_cluster,
#                 params={"id_project": id_project, "id_workspace": id_workspace},
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         # logger.debug("url: {}".format(r.url))
#         # logger.debug("status: {}".format(r.status_code))
#         # logger.debug("json: {}".format(r.json()))
#
#         gke_cluster_dict: Union[Dict[Any, Any], List[Any]] = r.json()
#
#         if gke_cluster_dict is None:
#             return None
#         if not isinstance(gke_cluster_dict, dict):
#             raise exceptions.ZeusGcpApiException(
#                 "Could not parse GKEClusterSchema response"
#             )
#
#         # If the request failed, display the message from ApiResponseData
#         if (
#             r.status_code == httpx.codes.BAD_REQUEST
#             and gke_cluster_dict.get("status", "FAIL") == "FAIL"
#         ):
#             logger.debug(
#                 "Error while getting GKEClusterSchema: {}".format(
#                     gke_cluster_dict.get(
#                         "message", gke_cluster_dict.get("detail", "FAIL")
#                     )
#                 )
#             )
#
#         # If everything is good, convert gke_cluster_response to GKEClusterSchema and return
#         if r.status_code == httpx.codes.OK:
#             return schemas.GKEClusterSchema.from_dict(gke_cluster_dict)
#
#     return None
#
#
# def get_or_create_gke_cluster_for_gcp_project(
#     id_workspace: int, id_project: str, pak8_conf: Pak8Conf
# ) -> Optional[schemas.GKEClusterRelease]:
#
#     logger.debug(
#         "Creating GKEClusterSchema for workspace {} and project {}".format(
#             id_workspace, id_project
#         )
#     )
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             # Set the timeout to None because the response for this request could take a while
#             r: httpx.Response = api.post(
#                 ZeusApi.get_or_create_gke_cluster,
#                 params={"id_project": id_project, "id_workspace": id_workspace},
#                 data=pak8_conf.json(),
#                 timeout=None,
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_error(
#                 "NetworkError, please confirm you are connected to the internet"
#             )
#             return None
#
#         logger.debug("url: {}".format(r.url))
#         logger.debug("status: {}".format(r.status_code))
#         logger.debug("json: {}".format(r.json()))
#
#         gke_cluster_release_dict: Union[Dict[Any, Any], List[Any]] = r.json()
#
#         if gke_cluster_release_dict is None:
#             return None
#         if not isinstance(gke_cluster_release_dict, dict):
#             raise exceptions.ZeusGcpApiException(
#                 "Could not parse GKEClusterRelease response"
#             )
#
#         # If the request failed, display the message from ApiResponseData
#         if (
#             r.status_code == httpx.codes.BAD_REQUEST
#             and gke_cluster_release_dict.get("status", "FAIL") == "FAIL"
#         ):
#             pprint_error(
#                 gke_cluster_release_dict.get(
#                     "message", gke_cluster_release_dict.get("detail", "FAIL")
#                 )
#             )
#
#         # If everything is good, convert gke_cluster_response to GKEClusterSchema and return
#         if r.status_code == httpx.codes.OK:
#             return schemas.GKEClusterRelease.from_dict(gke_cluster_release_dict)
#
#     return None
#
#
# def delete_gke_cluster_if_exists(
#     id_workspace: int, id_project: str
# ) -> Optional[schemas.GKEClusterRelease]:
#
#     logger.debug(
#         "Deleting GKEClusterSchema for workspace {} and project {}".format(
#             id_workspace, id_project
#         )
#     )
#     authenticated_client = get_authenticated_client()
#     if authenticated_client is None:
#         return None
#     with authenticated_client as api:
#         try:
#             # Set the timeout to None because the response for this request could take a while
#             r: httpx.Response = api.get(
#                 ZeusApi.delete_gke_cluster_if_exists,
#                 params={"id_project": id_project, "id_workspace": id_workspace},
#                 timeout=None,
#             )
#             validate_response(r)
#         except httpx.NetworkError as e:
#             pprint_network_error()
#             return None
#
#         logger.debug("url: {}".format(r.url))
#         logger.debug("status: {}".format(r.status_code))
#         logger.debug("json: {}".format(r.json()))
#
#         gke_cluster_release_dict: Union[Dict[Any, Any], List[Any]] = r.json()
#
#         if gke_cluster_release_dict is None:
#             return None
#         if not isinstance(gke_cluster_release_dict, dict):
#             raise exceptions.ZeusGcpApiException(
#                 "Could not parse GKEClusterRelease response"
#             )
#
#         # If the request failed, display the message from ApiResponseData
#         if (
#             r.status_code == httpx.codes.BAD_REQUEST
#             and gke_cluster_release_dict.get("status", "FAIL") == "FAIL"
#         ):
#             pprint_error(
#                 gke_cluster_release_dict.get(
#                     "message", gke_cluster_release_dict.get("detail", "FAIL")
#                 )
#             )
#
#         # If everything is good, convert gke_cluster_response to GKEClusterRelease and return
#         if r.status_code == httpx.codes.OK:
#             return schemas.GKEClusterRelease.from_dict(gke_cluster_release_dict)
#
#     return None



from typing import Dict, Optional

import typer
from httpx import Client, Response, codes

from phiterm.conf.constants import (
    APP_NAME,
    APP_VERSION,
    BACKEND_API_URL,
    PHI_AUTH_TOKEN_COOKIE,
)
from phiterm.conf.auth import read_auth_token
from phiterm.utils.cli_console import print_error


class ZeusApi:
    # Auth paths
    authenticate: str = "auth/authenticate"
    cli_auth: str = "auth/cli-auth"
    # WorkspaceSchema paths
    upsert_workspace: str = "workspaces/upsertfromcli"
    update_workspace: str = "workspaces/update"
    read_primary_workspace: str = "workspaces/read/primary"
    read_available_workspaces: str = "workspaces/read/available"
    deploy_workspace: str = "workspaces/deploy"
    shutdown_workspace: str = "workspaces/shutdown"
    sync_pak8_conf: str = "workspaces/syncpak8conf"
    # GCP paths
    upsert_gcp_project: str = "gcp/upsertfromcli"
    update_gcp_project: str = "gcp/update/gcpproject"
    read_primary_gcp_project: str = "gcp/read/primarygcpproject"
    get_gke_cluster: str = "gcp/gke/getcluster"
    get_or_create_gke_cluster: str = "gcp/gke/getorcreatecluster"
    delete_gke_cluster_if_exists: str = "gcp/gke/deleteclusterifexists"
    # Kubectl paths
    get_k8s_manifests: str = "kubectl/get/manifests"
    get_active_resources: str = "kubectl/get/active"
    apply_k8s_manifests: str = "kubectl/apply/manifests"


paths: Dict[str, str] = {
    # Auth paths
    "authenticate": "auth/authenticate",
    "cli_auth": "auth/cli-auth",
    # WorkspaceSchema paths
    "upsert_workspace": "workspaces/upsertfromcli",
    "update_workspace": "workspaces/update",
    "read_primary_workspace": "workspaces/read/primary",
    "read_available_workspaces": "workspaces/read/available",
    "deploy_workspace": "workspaces/deploy",
    "shutdown_workspace": "workspaces/shutdown",
    "sync_pak8_conf": "workspaces/syncpak8conf",
    # GCP paths
    "upsert_gcp_project": "gcp/upsertfromcli",
    "update_gcp_project": "gcp/update/gcpproject",
    "read_active_gcp_project": "gcp/read/activegcpproject",
    "get_gke_cluster": "gcp/gke/getcluster",
    "get_or_create_gke_cluster": "gcp/gke/getorcreatecluster",
    "delete_gke_cluster_if_exists": "gcp/gke/deleteclusterifexists",
    # Kubectl paths
    "get_k8s_manifests": "kubectl/get/manifests",
    "get_active_resources": "kubectl/get/active",
    "apply_k8s_manifests": "kubectl/apply/manifests",
}

headers = {"user-agent": f"{APP_NAME}/{APP_VERSION}"}


def validate_response(r: Response) -> None:
    """Exits the application if the response is invalid
    Use it like:
        r: Response = api.get(paths["authenticate"])
        validate_response(r)
    """
    # Check status code:
    if r.status_code == codes.UNAUTHORIZED:
        print_error(
            "There seems to be an authentication issue, please run `phi init` again"
        )
        raise typer.Exit()


def get_authenticated_client() -> Optional[Client]:
    # Returns an instance of httpx.Client which with preconfigured auth and base url.
    try:
        session_cookie: Optional[str] = read_auth_token()
    except Exception:
        # print_error("Could not authenticating user. Please run `phi auth` to fix")
        return None

    if session_cookie is None:
        return None
    return Client(
        base_url=BACKEND_API_URL,
        headers=headers,
        cookies={PHI_AUTH_TOKEN_COOKIE: session_cookie},
        timeout=60,
    )



from typing import List, Optional

from pydantic import BaseModel

######################################################
# K8s Schemas
# These should match ~/philab/services/zeus/zeus/schemas/k8s_schemas.py
######################################################


class K8sFilters(BaseModel):
    type_filters: Optional[List[str]] = None
    name_filters: Optional[List[str]] = None



from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from phiterm.enums.user import UserPermissions


class UserSchema(BaseModel):
    """
    Schema used for user login and registration.
    """

    email: str
    id_user: Optional[int] = None
    handle: Optional[str] = None
    is_active: Optional[bool] = True
    is_admin: Optional[bool] = False
    is_bot: Optional[bool] = False
    is_test: Optional[bool] = False
    name: Optional[str] = None
    locale: Optional[str] = None
    country_code: Optional[str] = None
    tos_accepted: Optional[bool] = False
    email_verified: Optional[bool] = False
    user_permissions: Optional[List[UserPermissions]] = None

    @classmethod
    def from_dict(cls, user_dict: Dict[str, Any]):
        return cls(
            email=user_dict.get("email"),
            id_user=user_dict.get("id_user"),
            handle=user_dict.get("handle"),
            is_active=user_dict.get("is_active"),
            is_admin=user_dict.get("is_admin"),
            is_bot=user_dict.get("is_bot"),
            is_test=user_dict.get("is_test"),
            name=user_dict.get("name"),
            locale=user_dict.get("locale"),
            country_code=user_dict.get("country_code"),
            tos_accepted=user_dict.get("tos_accepted"),
            email_verified=user_dict.get("email_verified"),
            user_permissions=user_dict.get("user_permissions"),
        )


class EmailPasswordSignInSchema(BaseModel):
    email: str
    password: str
    auth_source: str = "cli"



from dataclasses import dataclass
from typing import Optional


@dataclass
class ApiRoutes:

    # user paths
    USER_PING: str = "v1/user/ping"
    USER_READ: str = "v1/user/read"
    USER_UPDATE: str = "v1/user/update"
    USER_SIGN_IN: str = "v1/user/signin"
    USER_CLI_AUTH: str = "v1/user/cliauth"
    USER_AUTHENTICATE: str = "v1/user/authenticate"
    USER_AUTH_REFRESH: str = "v1/user/authrefresh"

    # workspace paths
    WORKSPACE_PING: str = "v1/workspace/ping"
    WORKSPACE_CREATE: str = "v1/workspace/create"
    WORKSPACE_UPDATE: str = "v1/workspace/update"
    WORKSPACE_UPDATE_PRIMARY: str = "v1/workspace/update/primary"
    WORKSPACE_READ_PRIMARY: str = "v1/workspace/read/primary"
    WORKSPACES_READ_AVAILABLE: str = "v1/workspace/read/available"



from typing import Optional, List

from phiterm.conf.phi_conf import PhiWsData
from phiterm.utils.cli_console import (
    print_info,
    print_heading,
    print_info,
)
from phiterm.utils.log import logger
from phiterm.workspace.ws_enums import WorkspaceEnv
from phiterm.workspace.ws_operator import filter_and_prep_configs

from phidata.infra.base import InfraConfig


def run_command(
    command: str,
    ws_data: PhiWsData,
    target_env: str,
    target_app: Optional[str] = None,
) -> None:
    """Run a command in databox."""
    from phidata.infra.docker.config import DockerConfig
    from phidata.infra.k8s.config import K8sConfig

    if ws_data is None or ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return
    ws_config = ws_data.ws_config
    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    # Get the config to run the command in
    filtered_configs: List[InfraConfig] = filter_and_prep_configs(
        ws_config=ws_config,
        target_env=target_env,
    )

    # Final run status
    run_status_list: List[bool] = []
    for config in filtered_configs:
        if isinstance(config, DockerConfig):
            from phiterm.docker.docker_operator import run_command_docker

            _run_status = run_command_docker(
                command=command,
                docker_config=config,
                target_app=target_app,
            )
            run_status_list.append(_run_status)
        if isinstance(config, K8sConfig):
            from phiterm.k8s.k8s_operator import run_command_k8s

            _run_status = run_command_k8s(
                command=command.split(),
                k8s_config=config,
                target_app=target_app,
            )
            run_status_list.append(_run_status)

    if all(run_status_list):
        print_heading("Command run success")
    else:
        logger.error(f"Command run failure: {run_status_list}")



# import datetime
# from typing import Any, Dict, Optional
#
# from pydantic import BaseModel
#
# import pak8.gcp.enums as pak8_gcp_enums
# from phiterm.gcp.gcp_enums import GCPProjectStatus
# from phiterm.utils.dttm import dttm_str_to_dttm
#
# ######################################################
# # GCP Schemas
# # These should match ~/philab/services/zeus/zeus/schemas/gcp_schemas.py
# ######################################################
#
#
# class GCPProjectSchema(BaseModel):
#     name: str
#     is_active: bool
#     is_verified: bool = False
#     is_test: bool = False
#     gcp_project_id: str
#     id_project: str
#     id_workspace: int
#     created_by_id_user: int
#     created_by_email: str
#     create_ts: datetime.datetime
#     gcp_zone: Optional[str] = None
#     gcp_service_account_key_available: bool = False
#     vpc_network: Optional[str]
#     last_update_ts: Optional[datetime.datetime] = None
#     project_status: Optional[GCPProjectStatus]
#
#     @classmethod
#     def from_dict(cls, gcp_project_dict: Dict[str, Any]):
#         return cls(
#             name=gcp_project_dict.get("name", None),
#             is_active=gcp_project_dict.get("is_valid", None),
#             is_verified=gcp_project_dict.get("is_verified", None),
#             is_test=gcp_project_dict.get("is_test", None),
#             gcp_project_id=gcp_project_dict.get("gcp_project_id", None),
#             id_project=gcp_project_dict.get("id_project", None),
#             id_workspace=gcp_project_dict.get("id_workspace", None),
#             created_by_id_user=gcp_project_dict.get("created_by_id_user", None),
#             created_by_email=gcp_project_dict.get("created_by_email", None),
#             create_ts=dttm_str_to_dttm(gcp_project_dict.get("create_ts", None)),
#             gcp_zone=gcp_project_dict.get("gcp_zone", None),
#             gcp_service_account_key_available=gcp_project_dict.get(
#                 "gcp_service_account_key_available", False
#             ),
#             vpc_network=gcp_project_dict.get("vpc_network", None),
#             last_update_ts=dttm_str_to_dttm(
#                 gcp_project_dict.get("last_update_ts", None)
#             ),
#             project_status=GCPProjectStatus.from_str(
#                 gcp_project_dict.get("project_status", None)
#             ),
#         )
#
#     # def secho_status(self):
#     #     pprint_info("")
#     #     pprint_subheading("GCP Project: {}".format(self.gcp_project_id))
#     #     pprint_status("Active: {}".format(self.is_valid))
#     #     pprint_status(
#     #         "Status: {}".format(
#     #             self.project_status.value if self.project_status else "unavailable"
#     #         )
#     #     )
#     #     pprint_status("name: {}".format(gcp_project.name))
#     #     pprint_status("id_project: {}".format(gcp_project.id_project))
#     #     pprint_status("id_workspace: {}".format(gcp_project.id_workspace))
#     #     pprint_status("gcp_service_account_key_available: {}".format(gcp_project.gcp_service_account_key_available))
#
#
# class UpsertGCPProjectFromCli(BaseModel):
#     # Required
#     name: str
#     id_project: str
#     id_workspace: int
#     gcp_project_id: str
#     gcp_service_account_key: Dict[str, Any]
#     # Optional
#     gcp_service_account: Optional[Dict[str, Any]] = None
#     gcp_zone: Optional[str] = None
#     vpc_network: Optional[str] = None
#     is_active: bool = True
#     is_verified: bool = False
#     is_test: bool = False
#     set_as_primary_project_for_ws: bool = False
#
#
# class UpdateGCPProject(BaseModel):
#     id_project: str
#     id_workspace: int
#     name: Optional[str]
#     is_active: Optional[bool]
#     is_verified: Optional[bool]
#     is_test: Optional[bool]
#     gcp_zone: Optional[str]
#     gcp_project_id: Optional[str]
#     gcp_service_account_key: Optional[Dict[str, Any]]
#     vpc_network: Optional[str]
#     set_as_primary_project_for_ws: Optional[bool]
#
#
# class GKEClusterSchema(BaseModel):
#     is_active: bool
#     is_verified: bool = False
#     is_test: bool = False
#     id_cluster: str
#     id_project: str
#     id_workspace: int
#     created_by_id_user: int
#     created_by_email: str
#     create_ts: datetime.datetime
#     gcp_zone: Optional[str]
#     gcp_project_id: str
#     cluster_name: str
#     current_cluster_status: pak8_gcp_enums.GKEClusterStatus
#     current_cluster_config: Optional[Dict[str, Any]] = None
#     last_update_ts: Optional[datetime.datetime] = None
#     deactivate_ts: Optional[datetime.datetime] = None
#
#     class Config:
#         arbitrary_types_allowed = True
#
#     @classmethod
#     def from_dict(cls, gke_cluster_dict: Dict[str, Any]):
#         return cls(
#             is_active=gke_cluster_dict.get("is_valid", None),
#             is_verified=gke_cluster_dict.get("is_verified", None),
#             is_test=gke_cluster_dict.get("is_test", None),
#             id_cluster=gke_cluster_dict.get("id_cluster", None),
#             id_project=gke_cluster_dict.get("id_project", None),
#             id_workspace=gke_cluster_dict.get("id_workspace", None),
#             created_by_id_user=gke_cluster_dict.get("created_by_id_user", None),
#             created_by_email=gke_cluster_dict.get("created_by_email", None),
#             create_ts=dttm_str_to_dttm(gke_cluster_dict.get("create_ts", None)),
#             gcp_zone=gke_cluster_dict.get("gcp_zone", None),
#             gcp_project_id=gke_cluster_dict.get("gcp_project_id", None),
#             cluster_name=gke_cluster_dict.get("cluster_name", None),
#             current_cluster_status=pak8_gcp_enums.GKEClusterStatus.from_str(
#                 gke_cluster_dict.get("current_cluster_status", None)
#             ),
#             current_cluster_config=gke_cluster_dict.get("current_cluster_config", None),
#             last_update_ts=dttm_str_to_dttm(
#                 gke_cluster_dict.get("last_update_ts", None)
#             ),
#             deactivate_ts=dttm_str_to_dttm(gke_cluster_dict.get("deactivate_ts", None)),
#         )
#
#     # def secho_status(self):
#     #     pprint_info("")
#     #     pprint_subheading("GKE Cluster: {}".format(self.cluster_name))
#     #     pprint_status("Active: {}".format(self.is_valid))
#     #     pprint_status("Status: {}".format(self.current_cluster_status.value))
#     #     pprint_status("Cluster config: {})".format(gke_cluster.current_cluster_config))
#
#
# class CreateGKECluster(BaseModel):
#     is_verified: bool = False
#     is_test: bool = False
#     id_cluster: str
#     id_project: str
#     id_workspace: int
#     created_by_id_user: int
#     created_by_email: str
#     create_ts: datetime.datetime
#     gcp_zone: Optional[str]
#     gcp_project_id: str
#     cluster_name: str
#
#     class Config:
#         arbitrary_types_allowed = True



"""Phi Workspace Cli

This is the entrypoint for the `phi ws` commands
"""
from typing import Optional, cast

import typer

from phiterm.utils.cli_console import (
    print_error,
    print_info,
    print_heading,
    print_available_workspaces,
    print_conf_not_available_msg,
    print_active_workspace_not_available,
)
from phiterm.utils.log import logger, set_log_level_to_debug
from phiterm.workspace.ws_enums import (
    WorkspaceEnv,
    WorkspaceStarterTemplate,
    WorkspaceConfigType,
)

ws_app = typer.Typer(
    name="ws",
    short_help="Manage workspaces",
    help="""\b
Use `phi ws <command>` to create, setup or update your workspace.
Run `phi ws <command> --help` for more info.
""",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="<command>",
)


@ws_app.command(short_help="Create a new workspace in the current directory.")
def init(
    ws_name: str = typer.Option(
        None, "-ws", help="Name of the new workspace", show_default=False
    ),
    template: str = typer.Option(
        "aws",
        "-t",
        "--template",
        help="Choose a starter template which comes pre-populated with defaults",
        show_default=True,
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Creates a new workspace in the current directory using the selected starter template
    [default: aws]

    \b
    Examples:
    $ phi ws init                 -> Create a new workspace
    $ phi ws init -ws data        -> Create a workspace named data using the default starter template (aws)
    """
    from phiterm.workspace.ws_operator import (
        create_new_workspace,
        initialize_workspace,
    )

    if print_debug_log:
        set_log_level_to_debug()

    if ws_name is None:
        initialize_workspace()
        return

    if (
        template is None
        or template.lower() not in WorkspaceStarterTemplate.values_list()
    ):
        print_error(
            f"{template} is not a supported template, please choose from: {WorkspaceStarterTemplate.values_list()}"
        )
        return
    _template: WorkspaceStarterTemplate = cast(
        WorkspaceStarterTemplate,
        WorkspaceStarterTemplate.from_str(template),
    )

    create_new_workspace(ws_name, _template)


@ws_app.command(short_help="Setup phidata workspace from the current directory")
def setup(
    ws_name: str = typer.Option(None, "-ws", help="Name of the workspace to setup"),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Setup a phidata workspace from the current directory.
    This command can be run from within the workspace directory
        OR with a "-ws" flag to setup a workspace by name.

    \b
    Examples:
    $ `phi ws setup`           -> Setup the current directory as a phidata workspace
    $ `phi ws setup -ws idata` -> Setup the workspace named idata
    """
    from pathlib import Path
    from phiterm.workspace.ws_operator import setup_workspace

    if print_debug_log:
        set_log_level_to_debug()

    # By default, we assume this command is run from the workspace directory
    if ws_name is None:
        # If the user does not provide a ws_name, that implies `phi ws setup` is ran from
        # the workspace directory.
        ws_path: Path = Path(".").resolve()
        setup_workspace(ws_path)
    else:
        # If the user provides a workspace name manually, we find the dir for that ws and set it up
        from phiterm.conf.phi_conf import PhiConf

        phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
        if not phi_conf:
            print_conf_not_available_msg()
            return

        _ws_path: Optional[Path] = phi_conf.get_ws_root_path_by_name(ws_name)
        if _ws_path is None:
            print_error(f"Could not find workspace {ws_name}")
            avl_ws = phi_conf.available_ws
            if avl_ws:
                print_available_workspaces(avl_ws)
            return
        setup_workspace(_ws_path)


@ws_app.command(short_help="Create resources for active workspace")
def up(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:CONFIG:GROUP:NAME:TYPE",
        metavar="[FILTER]",
    ),
    env_filter: Optional[str] = typer.Option(
        None, "-e", "--env", metavar="", help="Filter the environment to deploy"
    ),
    config_filter: Optional[str] = typer.Option(
        None, "-c", "--config", metavar="", help="Filter the config to deploy"
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter using group name"
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be deployed and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before deploying resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Force",
    ),
):
    """\b
    Create resources for the active workspace.
    Options can be used to limit the resources to deploy.
      --env     : Env (dev, stg, prd)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
      --config  : Config type (docker, aws, k8s)
    \b
    The Filter can also be provided as a single argument as - ENV:CONFIG:APP:NAME:TYPE
    \b
    Examples:
    $ `phi ws up`            -> Deploy all resources
    $ `phi ws up dev`        -> Deploy all dev resources
    $ `phi ws up prd`        -> Deploy all prd resources
    $ `phi ws up prd:aws`    -> Deploy all prd aws resources
    $ `phi ws up prd:::s3`   -> Deploy prd resources matching name s3
    """
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.workspace.ws_operator import deploy_workspace
    from phiterm.utils.load_env import load_env
    from phiterm.utils.ws_filter import parse_ws_filter

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(
        env={
            "PHI_WS_FORCE": str(force),
        },
        dotenv_dir=active_ws_data.ws_root_path,
    )

    target_config: Optional[WorkspaceConfigType] = None
    target_config_str: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None
    target_app: Optional[str] = None
    target_env: Optional[str] = None

    # derive env/name/type/config from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(
                f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}"
            )
        (
            target_env,
            target_config_str,
            target_app,
            target_name,
            target_type,
        ) = parse_ws_filter(resource_filter)

    # derive env/app/name/type/config from command options
    if (
        target_config_str is None
        and config_filter is not None
        and isinstance(config_filter, str)
    ):
        target_config_str = config_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter
    if (
        target_app is None
        and group_filter is not None
        and isinstance(group_filter, str)
    ):
        target_app = group_filter
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter

    # derive env/config/name/type from defaults
    if target_env is None:
        target_env = (
            active_ws_data.ws_config.default_env if active_ws_data.ws_config else None
        )
    if target_config_str is None:
        target_config_str = (
            active_ws_data.ws_config.default_config
            if active_ws_data.ws_config
            else None
        )
    if target_config_str is not None:
        if target_config_str.lower() not in WorkspaceConfigType.values_list():
            print_error(
                f"{target_config_str} is not supported, please choose from: {WorkspaceConfigType.values_list()}"
            )
            return
        target_config = cast(
            WorkspaceConfigType,
            WorkspaceConfigType.from_str(target_config_str),
        )

    logger.debug("Deploying workspace")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_config: {target_config}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_app   : {target_app}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    print_heading("Deploying workspace: {}\n".format(active_ws_data.ws_name))
    deploy_workspace(
        ws_data=active_ws_data,
        target_env=target_env,
        target_config=target_config,
        target_name=target_name,
        target_type=target_type,
        target_app=target_app,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
    )


@ws_app.command(short_help="Delete resources for active workspace")
def down(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:CONFIG:GROUP:NAME:TYPE",
        metavar="[FILTER]",
    ),
    env_filter: str = typer.Option(
        None, "-e", "--env", metavar="", help="Filter the environment to shut down"
    ),
    config_filter: str = typer.Option(
        None, "-c", "--config", metavar="", help="Filter the config to shut down"
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter using group name"
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be deleted and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before deleting resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Force",
    ),
):
    """\b
    Delete resources for active workspace.
    Options can be used to limit the resources to delete.
      --env     : Env (dev, stg, prd)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
      --config  : Config type (docker, aws, k8s)
    \b
    The Filter can also be provided as a single argument as - ENV:CONFIG:APP:NAME:TYPE
    \b
    Examples:
    $ `phi ws down`            -> Delete all resources
    """
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.workspace.ws_operator import shutdown_workspace
    from phiterm.utils.load_env import load_env
    from phiterm.utils.ws_filter import parse_ws_filter

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(
        env={
            "PHI_WS_FORCE": str(force),
        },
        dotenv_dir=active_ws_data.ws_root_path,
    )

    target_config: Optional[WorkspaceConfigType] = None
    target_config_str: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None
    target_app: Optional[str] = None
    target_env: Optional[str] = None

    # derive env/config/name/type from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(
                f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}"
            )
        (
            target_env,
            target_config_str,
            target_app,
            target_name,
            target_type,
        ) = parse_ws_filter(resource_filter)

    # derive env/app/name/type/config from command options
    if (
        target_config_str is None
        and config_filter is not None
        and isinstance(config_filter, str)
    ):
        target_config_str = config_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter
    if (
        target_app is None
        and group_filter is not None
        and isinstance(group_filter, str)
    ):
        target_app = group_filter
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter

    # derive env/config/name/type from defaults
    if target_env is None:
        target_env = (
            active_ws_data.ws_config.default_env if active_ws_data.ws_config else None
        )
    if target_config_str is None:
        target_config_str = (
            active_ws_data.ws_config.default_config
            if active_ws_data.ws_config
            else None
        )
    if target_config_str is not None:
        if target_config_str.lower() not in WorkspaceConfigType.values_list():
            print_error(
                f"{target_config_str} is not supported, please choose from: {WorkspaceConfigType.values_list()}"
            )
            return
        target_config = cast(
            WorkspaceConfigType,
            WorkspaceConfigType.from_str(target_config_str),
        )

    logger.debug("Shutting down workspace")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_config: {target_config}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_app   : {target_app}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    print_heading("Shutdown workspace: {}\n".format(active_ws_data.ws_name))
    shutdown_workspace(
        ws_data=active_ws_data,
        target_env=target_env,
        target_config=target_config,
        target_name=target_name,
        target_type=target_type,
        target_app=target_app,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
    )


@ws_app.command(short_help="Update resources for active workspace")
def patch(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:CONFIG:GROUP:NAME:TYPE",
        metavar="[FILTER]",
    ),
    env_filter: str = typer.Option(
        None, "-e", "--env", metavar="", help="Filter the environment to patch"
    ),
    config_filter: str = typer.Option(
        None, "-c", "--config", metavar="", help="Filter the config to patch"
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter using group name"
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be patched and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before patching resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Force",
    ),
):
    """\b
    Update resources for active workspace.
    Options can be used to limit the resources to update.
      --env     : Env (dev, stg, prd)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
      --config  : Config type (docker, aws, k8s)
    \b
    The Filter can also be provided as a single argument as - ENV:CONFIG:APP:NAME:TYPE
    Examples:
    \b
    $ `phi ws patch`           -> Patch all resources
    """
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.workspace.ws_operator import patch_workspace
    from phiterm.utils.load_env import load_env
    from phiterm.utils.ws_filter import parse_ws_filter

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(
        env={
            "PHI_WS_FORCE": str(force),
        },
        dotenv_dir=active_ws_data.ws_root_path,
    )

    target_config: Optional[WorkspaceConfigType] = None
    target_config_str: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None
    target_app: Optional[str] = None
    target_env: Optional[str] = None

    # derive env/name/type/config from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(
                f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}"
            )
        (
            target_env,
            target_config_str,
            target_app,
            target_name,
            target_type,
        ) = parse_ws_filter(resource_filter)

    # derive env/app/name/type/config from command options
    if (
        target_config_str is None
        and config_filter is not None
        and isinstance(config_filter, str)
    ):
        target_config_str = config_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter
    if (
        target_app is None
        and group_filter is not None
        and isinstance(group_filter, str)
    ):
        target_app = group_filter
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter

    # derive env/config/name/type from defaults
    if target_env is None:
        target_env = (
            active_ws_data.ws_config.default_env if active_ws_data.ws_config else None
        )
    if target_config_str is None:
        target_config_str = (
            active_ws_data.ws_config.default_config
            if active_ws_data.ws_config
            else None
        )
    if target_config_str is not None:
        if target_config_str.lower() not in WorkspaceConfigType.values_list():
            print_error(
                f"{target_config_str} is not supported, please choose from: {WorkspaceConfigType.values_list()}"
            )
            return
        target_config = cast(
            WorkspaceConfigType,
            WorkspaceConfigType.from_str(target_config_str),
        )

    logger.debug("Patching workspace")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_config: {target_config}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_app   : {target_app}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    print_heading("Patching workspace: {}\n".format(active_ws_data.ws_name))
    patch_workspace(
        ws_data=active_ws_data,
        target_env=target_env,
        target_config=target_config,
        target_name=target_name,
        target_type=target_type,
        target_app=target_app,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
    )


@ws_app.command(short_help="Restart resources for active workspace")
def restart(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:CONFIG:GROUP:NAME:TYPE",
        metavar="[FILTER]",
    ),
    env_filter: str = typer.Option(
        None, "-e", "--env", metavar="", help="Filter the environment to restart"
    ),
    config_filter: str = typer.Option(
        None, "-i", "--config", metavar="", help="Filter the config to restart"
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter using group name"
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be restarted and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before restarting resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Force",
    ),
):
    """\b
    Restarts the active workspace. i.e. runs `phi ws down` and then `phi ws up`.

    \b
    Examples:
    $ `phi ws restart`
    """
    from time import sleep

    down(
        resource_filter=resource_filter,
        env_filter=env_filter,
        config_filter=config_filter,
        name_filter=name_filter,
        type_filter=type_filter,
        group_filter=group_filter,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        print_debug_log=print_debug_log,
        force=force,
    )
    print_info("Sleeping for 2 seconds..")
    sleep(2)
    up(
        resource_filter=resource_filter,
        env_filter=env_filter,
        config_filter=config_filter,
        name_filter=name_filter,
        type_filter=type_filter,
        group_filter=group_filter,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        print_debug_log=print_debug_log,
        force=force,
    )


@ws_app.command(short_help="Print workspace status")
def status(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:CONFIG:GROUP:NAME:TYPE",
        metavar="[FILTER]",
    ),
    env_filter: str = typer.Option(
        None, "-e", "--env", metavar="", help="Filter the environment"
    ),
    config_filter: str = typer.Option(
        None, "-c", "--config", metavar="", help="Filter the config"
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter using group name"
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be displayed and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before displaying resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        False,
        "-f",
        "--force",
        help="Force",
    ),
):
    """\b
    Get resource status for active workspace.
    Options can be used to limit the resources to update.
      --env     : Env (dev, stg, prd)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
      --config  : Config type (docker, aws, k8s)
    \b
    The Filter can also be provided as a single argument as - ENV:CONFIG:APP:NAME:TYPE
    Examples:
    \b
    $ `phi ws patch`           -> Patch all resources
    """
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.workspace.ws_operator import print_workspace_status
    from phiterm.utils.load_env import load_env
    from phiterm.utils.ws_filter import parse_ws_filter

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(
        env={
            "PHI_WS_FORCE": str(force),
        },
        dotenv_dir=active_ws_data.ws_root_path,
    )

    target_config: Optional[WorkspaceConfigType] = None
    target_config_str: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None
    target_app: Optional[str] = None
    target_env: Optional[str] = None

    # derive env/name/type/config from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(
                f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}"
            )
        (
            target_env,
            target_config_str,
            target_app,
            target_name,
            target_type,
        ) = parse_ws_filter(resource_filter)

    # derive env/app/name/type/config from command options
    if (
        target_config_str is None
        and config_filter is not None
        and isinstance(config_filter, str)
    ):
        target_config_str = config_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter
    if (
        target_app is None
        and group_filter is not None
        and isinstance(group_filter, str)
    ):
        target_app = group_filter
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter

    # derive env/config/name/type from defaults
    if target_env is None:
        target_env = (
            active_ws_data.ws_config.default_env if active_ws_data.ws_config else None
        )
    if target_config_str is None:
        target_config_str = (
            active_ws_data.ws_config.default_config
            if active_ws_data.ws_config
            else None
        )
    if target_config_str is not None:
        if target_config_str.lower() not in WorkspaceConfigType.values_list():
            print_error(
                f"{target_config_str} is not supported, please choose from: {WorkspaceConfigType.values_list()}"
            )
            return
        target_config = cast(
            WorkspaceConfigType,
            WorkspaceConfigType.from_str(target_config_str),
        )

    logger.debug("Reading workspace status")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_config: {target_config}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_app   : {target_app}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    print_heading("Reading workspace status: {}\n".format(active_ws_data.ws_name))
    print_workspace_status(
        ws_data=active_ws_data,
        target_env=target_env,
        target_config=target_config,
        target_name=target_name,
        target_type=target_type,
        target_app=target_app,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
    )


@ws_app.command(short_help="Prints active workspace config", hidden=True)
def config(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Prints the active workspace config

    \b
    Examples:
    $ `phi ws config`         -> Print the active workspace config
    """
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.utils.load_env import load_env

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(
        dotenv_dir=active_ws_data.ws_root_path,
    )
    active_ws_data.print_to_cli()


# @ws_app.command(short_help="Show steps to setup a git repo for the ws")
# def git(
#     ws_name: str = typer.Option(None, "-ws", help="[Optional] Name for the workspace"),
#     print_debug_log: bool = typer.Option(
#         False,
#         "-d",
#         "--debug",
#         help="Print debug logs.",
#     ),
# ):
#     """
#     Print steps to setup a remote git repo for the active workspace.
#
#     \b
#     Examples:
#     $ `phi ws git`
#     $ `phi ws git -ws data`
#     """
#     from pathlib import Path
#     from phiterm.enums.user_enums import VersionControlProviderEnum
#     from phiterm.workspace.ws_operator import print_git_setup
#
#     from phiterm.conf.phi_conf import PhiConf
#
#     phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not phi_conf:
#         print_conf_not_available_msg()
#         return
#
#     _ws_name: Optional[str] = None
#     ws_root_path: Optional[Path] = None
#     if ws_name:
#         _ws_name = ws_name
#         ws_root_path = phi_conf.get_ws_root_path_by_name(_ws_name)
#     else:
#         # print steps for the active workspace
#         _ws_name = phi_conf.active_ws_name
#         if _ws_name is None:
#             print_info(
#                 "Primary workspace not available, searching current directory for a workspace"
#             )
#             ws_root_path = Path(".").resolve()
#             _ws_name = phi_conf.get_ws_name_by_path(ws_root_path)
#         else:
#             ws_data = phi_conf.get_ws_data_by_name(_ws_name)
#             if ws_data is not None and ws_data.ws_root_path is not None:
#                 ws_root_path = ws_data.ws_root_path
#
#     if _ws_name is None or ws_root_path is None:
#         print_error(f"Could not find workspace at {ws_root_path}")
#         avl_ws = phi_conf.available_ws
#         if avl_ws:
#             print_available_workspaces(avl_ws)
#         return
#
#     version_control_provider = (
#         phi_conf.user.version_control_provider
#         if phi_conf.user and phi_conf.user.version_control_provider
#         else VersionControlProviderEnum.GITHUB
#     )
#     print_git_setup(
#         ws_name=_ws_name,
#         ws_root_path=ws_root_path,
#         version_control_provider=version_control_provider,
#     )



from typing import Dict

from phiterm.workspace.ws_enums import WorkspaceStarterTemplate

template_to_repo_map: Dict[WorkspaceStarterTemplate, str] = {
    WorkspaceStarterTemplate.aws: "https://github.com/phidatahq/starter-aws.git",
    WorkspaceStarterTemplate.aws_snowflake: "https://github.com/phidatahq/starter-aws-snowflake.git",
    WorkspaceStarterTemplate.aws_duckdb: "https://github.com/phidatahq/starter-aws-duckdb.git",
    WorkspaceStarterTemplate.aws_backend: "https://github.com/phidatahq/starter-aws-backend.git",
}
